from __future__ import annotations

import logging
import os
from collections import defaultdict
from typing import Iterable, Sequence

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse
from fastapi.routing import APIRoute

from svc_infra.api.fastapi.docs.landing import CardSpec, DocTargets, render_index_html
from svc_infra.api.fastapi.docs.scoped import DOC_SCOPES
from svc_infra.api.fastapi.middleware.errors.catchall import CatchAllExceptionMiddleware
from svc_infra.api.fastapi.middleware.errors.handlers import register_error_handlers
from svc_infra.api.fastapi.middleware.idempotency import IdempotencyMiddleware
from svc_infra.api.fastapi.middleware.ratelimit import SimpleRateLimitMiddleware
from svc_infra.api.fastapi.middleware.request_id import RequestIdMiddleware
from svc_infra.api.fastapi.openapi.models import APIVersionSpec, ServiceInfo
from svc_infra.api.fastapi.openapi.mutators import setup_mutators
from svc_infra.api.fastapi.openapi.pipeline import apply_mutators
from svc_infra.api.fastapi.routers import register_all_routers
from svc_infra.app.env import CURRENT_ENVIRONMENT, DEV_ENV, LOCAL_ENV

logger = logging.getLogger(__name__)


def _gen_operation_id_factory():
    used: dict[str, int] = defaultdict(int)

    def _normalize(s: str) -> str:
        return "_".join(x for x in s.strip().replace(" ", "_").split("_") if x)

    def _gen(route: APIRoute) -> str:
        base = route.name or getattr(route.endpoint, "__name__", "op")
        base = _normalize(base)
        tag = _normalize(route.tags[0]) if route.tags else ""
        method = next(iter(route.methods or ["GET"])).lower()

        candidate = base
        if used[candidate]:
            if tag and not base.startswith(tag):
                candidate = f"{tag}_{base}"
            if used[candidate]:
                if not candidate.endswith(f"_{method}"):
                    candidate = f"{candidate}_{method}"
                if used[candidate]:
                    counter = used[candidate] + 1
                    candidate = f"{candidate}_{counter}"

        used[candidate] += 1
        return candidate

    return _gen


def _setup_cors(app: FastAPI, public_cors_origins: list[str] | str | None = None):
    if isinstance(public_cors_origins, list):
        origins = [o.strip() for o in public_cors_origins if o and o.strip()]
    elif isinstance(public_cors_origins, str):
        origins = [o.strip() for o in public_cors_origins.split(",") if o and o.strip()]
    else:
        fallback = os.getenv("CORS_ALLOW_ORIGINS", "http://localhost:3000")
        origins = [o.strip() for o in fallback.split(",") if o and o.strip()]

    if not origins:
        return

    cors_kwargs = dict(allow_credentials=True, allow_methods=["*"], allow_headers=["*"])
    if "*" in origins:
        cors_kwargs["allow_origin_regex"] = ".*"
    else:
        cors_kwargs["allow_origins"] = origins

    app.add_middleware(CORSMiddleware, **cors_kwargs)


def _setup_middlewares(app: FastAPI):
    app.add_middleware(RequestIdMiddleware)
    app.add_middleware(CatchAllExceptionMiddleware)
    app.add_middleware(IdempotencyMiddleware)
    app.add_middleware(SimpleRateLimitMiddleware)
    register_error_handlers(app)
    _add_route_logger(app)


def _coerce_list(value: str | Iterable[str] | None) -> list[str]:
    if value is None:
        return []
    if isinstance(value, str):
        return [value]
    return [v for v in value if v]


def _dump_or_none(model):
    return model.model_dump(exclude_none=True) if model is not None else None


def _build_child_app(service: ServiceInfo, spec: APIVersionSpec) -> FastAPI:
    title = f"{service.name} • {spec.tag}" if getattr(spec, "tag", None) else service.name
    child = FastAPI(
        title=title,
        version=service.release,
        contact=_dump_or_none(service.contact),
        license_info=_dump_or_none(service.license),
        terms_of_service=service.terms_of_service,
        description=service.description,
        generate_unique_id_function=_gen_operation_id_factory(),
    )

    _setup_middlewares(child)

    # ---- OpenAPI pipeline (DRY!) ----
    include_api_key = bool(spec.include_api_key) if spec.include_api_key is not None else False
    mount_path = f"/{spec.tag.strip('/')}"
    server_url = (
        mount_path
        if not spec.public_base_url
        else f"{spec.public_base_url.rstrip('/')}{mount_path}"
    )

    mutators = setup_mutators(
        service=service,
        spec=spec,
        include_api_key=include_api_key,
        server_url=server_url,
    )
    apply_mutators(child, *mutators)

    if spec.routers_package:
        register_all_routers(
            child, base_package=spec.routers_package, prefix="", environment=CURRENT_ENVIRONMENT
        )

    logger.info(
        "[%s] initialized version %s [env: %s]", service.name, spec.tag, CURRENT_ENVIRONMENT
    )
    return child


def _build_parent_app(
    service: ServiceInfo,
    *,
    public_cors_origins: list[str] | str | None,
    root_routers: list[str] | str | None,
    root_server_url: str | None = None,
    root_include_api_key: bool = False,
) -> FastAPI:
    show_root_docs = CURRENT_ENVIRONMENT in (LOCAL_ENV, DEV_ENV)

    parent = FastAPI(
        title=service.name,
        version=service.release,
        contact=_dump_or_none(service.contact),
        license_info=_dump_or_none(service.license),
        terms_of_service=service.terms_of_service,
        description=service.description,
        docs_url=("/docs" if show_root_docs else None),
        redoc_url=("/redoc" if show_root_docs else None),
        openapi_url=("/openapi.json" if show_root_docs else None),
    )

    _setup_cors(parent, public_cors_origins)
    _setup_middlewares(parent)

    mutators = setup_mutators(
        service=service,
        spec=None,
        include_api_key=root_include_api_key,
        server_url=root_server_url,
    )
    apply_mutators(parent, *mutators)

    # Root routers — svc-infra ping at '/', once
    register_all_routers(
        parent,
        base_package="svc_infra.api.fastapi.routers",
        prefix="",
        environment=CURRENT_ENVIRONMENT,
    )
    # app-provided root routers
    for pkg in _coerce_list(root_routers):
        register_all_routers(parent, base_package=pkg, prefix="", environment=CURRENT_ENVIRONMENT)

    return parent


def _add_route_logger(app: FastAPI):
    @app.middleware("http")
    async def _log_route(request, call_next):
        resp = await call_next(request)
        route = request.scope.get("route")
        # Prefer FastAPI's path_format (shows param patterns), fall back to path
        path = getattr(route, "path_format", None) or getattr(route, "path", None)
        if path:
            # Include mount root_path so mounted children show their full path
            root_path = request.scope.get("root_path", "") or ""
            resp.headers["X-Handled-By"] = f"{request.method} {root_path}{path}"
        return resp


def setup_service_api(
    *,
    service: ServiceInfo,
    versions: Sequence[APIVersionSpec],
    root_routers: list[str] | str | None = None,
    public_cors_origins: list[str] | str | None = None,
    root_public_base_url: str | None = None,
    root_include_api_key: bool | None = None,
) -> FastAPI:
    # infer if not explicitly provided
    effective_root_include_api_key = (
        any(bool(v.include_api_key) for v in versions)
        if root_include_api_key is None
        else bool(root_include_api_key)
    )

    root_server = root_public_base_url.rstrip("/") if root_public_base_url else "/"
    parent = _build_parent_app(
        service,
        public_cors_origins=public_cors_origins,
        root_routers=root_routers,
        root_server_url=root_server,
        root_include_api_key=effective_root_include_api_key,
    )

    # Mount each version
    for spec in versions:
        child = _build_child_app(service, spec)
        mount_path = f"/{spec.tag.strip('/')}"
        parent.mount(mount_path, child, name=spec.tag.strip("/"))

    @parent.get("/", include_in_schema=False)
    def index():
        cards: list[CardSpec] = []
        is_local_dev = CURRENT_ENVIRONMENT in (LOCAL_ENV, DEV_ENV)

        if is_local_dev:
            # Root card
            cards.append(
                CardSpec(
                    tag="",
                    docs=DocTargets(swagger="/docs", redoc="/redoc", openapi_json="/openapi.json"),
                )
            )

        # Version cards
        for spec in versions:
            tag = spec.tag.strip("/")
            cards.append(
                CardSpec(
                    tag=tag,
                    docs=DocTargets(
                        swagger=f"/{tag}/docs",
                        redoc=f"/{tag}/redoc",
                        openapi_json=f"/{tag}/openapi.json",
                    ),
                )
            )

        if is_local_dev:
            # Scoped cards (auth, payments, etc.)
            for scope, swagger, redoc, openapi_json, title in DOC_SCOPES:
                cards.append(
                    CardSpec(
                        tag=scope.strip("/"),
                        docs=DocTargets(swagger=swagger, redoc=redoc, openapi_json=openapi_json),
                    )
                )

        html = render_index_html(service_name=service.name, release=service.release, cards=cards)
        return HTMLResponse(html)

    return parent



from __future__ import annotations

from datetime import datetime
from typing import Optional

from sqlalchemy import JSON, Boolean, DateTime, Index, Numeric, String, UniqueConstraint, text
from sqlalchemy.orm import Mapped, mapped_column

from svc_infra.db.sql.authref import user_fk_constraint, user_id_type
from svc_infra.db.sql.base import ModelBase

TENANT_ID_LEN = 64


class PayCustomer(ModelBase):
    __tablename__ = "pay_customers"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    # Tenant scoping
    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    # Always typed to match the actual auth PK; FK is enforced at table level
    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_customer_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index("ix_pay_customers_user_provider", "user_id", "provider"),
        Index("ix_pay_customers_tenant_provider", "tenant_id", "provider"),
    )


class PayIntent(ModelBase):
    __tablename__ = "pay_intents"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_intent_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    amount: Mapped[int] = mapped_column(Numeric(18, 0), nullable=False)  # minor units
    currency: Mapped[str] = mapped_column(String(8), nullable=False)
    status: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    client_secret: Mapped[Optional[str]] = mapped_column(String(255))
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )
    confirmed_at: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))
    captured: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index("ix_pay_intents_user_provider", "user_id", "provider"),
        Index("ix_pay_intents_tenant_provider", "tenant_id", "provider"),
    )


class PayEvent(ModelBase):
    __tablename__ = "pay_events"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_event_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    type: Mapped[str] = mapped_column(String(64), index=True, nullable=False)
    received_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )
    payload_json: Mapped[dict] = mapped_column(JSON, nullable=False)  # compact JSON string


class LedgerEntry(ModelBase):
    __tablename__ = "ledger_entries"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    ts: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        server_default=text("CURRENT_TIMESTAMP"),
        nullable=False,
        index=True,
    )

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_ref: Mapped[Optional[str]] = mapped_column(String(128), index=True)
    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)
    amount: Mapped[int] = mapped_column(Numeric(18, 0), nullable=False)
    currency: Mapped[str] = mapped_column(String(8), nullable=False)
    kind: Mapped[str] = mapped_column(String(24), nullable=False)  # payment|refund|fee|payout...
    status: Mapped[str] = mapped_column(String(24), nullable=False)  # pending|posted|void

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index("ix_ledger_user_ts", "user_id", "ts"),
        Index("ix_ledger_tenant_provider", "tenant_id", "provider"),
        UniqueConstraint(
            "tenant_id",
            "provider",
            "provider_ref",
            "kind",
            name="uq_ledger_tenant_provider_ref_kind",
        ),
    )


class PayPaymentMethod(ModelBase):
    __tablename__ = "pay_methods"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)
    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_customer_id: Mapped[str] = mapped_column(String(128), index=True, nullable=False)
    provider_method_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    brand: Mapped[Optional[str]] = mapped_column(String(32))
    last4: Mapped[Optional[str]] = mapped_column(String(8))
    exp_month: Mapped[Optional[int]] = mapped_column(Numeric(2, 0))
    exp_year: Mapped[Optional[int]] = mapped_column(Numeric(4, 0))
    is_default: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index(
            "ix_pay_methods_tenant_provider_customer",
            "tenant_id",
            "provider",
            "provider_customer_id",
        ),
    )


class PayProduct(ModelBase):
    __tablename__ = "pay_products"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_product_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    name: Mapped[str] = mapped_column(String(128), nullable=False)
    active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )


class PayPrice(ModelBase):
    __tablename__ = "pay_prices"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_price_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    provider_product_id: Mapped[str] = mapped_column(String(128), index=True, nullable=False)
    currency: Mapped[str] = mapped_column(String(8), nullable=False)
    unit_amount: Mapped[int] = mapped_column(Numeric(18, 0), nullable=False)  # minor units
    interval: Mapped[Optional[str]] = mapped_column(String(16))  # month|year|week|day
    trial_days: Mapped[Optional[int]] = mapped_column(Numeric(5, 0))
    active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )


class PaySubscription(ModelBase):
    __tablename__ = "pay_subscriptions"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)
    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_customer_id: Mapped[str] = mapped_column(String(128), index=True, nullable=False)
    provider_subscription_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    provider_price_id: Mapped[str] = mapped_column(String(128), index=True, nullable=False)
    status: Mapped[str] = mapped_column(
        String(32), index=True, nullable=False
    )  # active|trialing|canceled|past_due|incomplete
    quantity: Mapped[int] = mapped_column(Numeric(10, 0), default=1, nullable=False)
    cancel_at_period_end: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)
    current_period_end: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index(
            "ix_pay_subscriptions_tenant_provider_customer",
            "tenant_id",
            "provider",
            "provider_customer_id",
        ),
    )


class PayInvoice(ModelBase):
    __tablename__ = "pay_invoices"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)
    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_invoice_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    provider_customer_id: Mapped[str] = mapped_column(String(128), index=True, nullable=False)
    status: Mapped[str] = mapped_column(
        String(24), index=True, nullable=False
    )  # draft|open|paid|void|uncollectible
    amount_due: Mapped[int] = mapped_column(Numeric(18, 0), nullable=False)
    currency: Mapped[str] = mapped_column(String(8), nullable=False)
    hosted_invoice_url: Mapped[Optional[str]] = mapped_column(String(255))
    pdf_url: Mapped[Optional[str]] = mapped_column(String(255))
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index(
            "ix_pay_invoices_tenant_provider_customer",
            "tenant_id",
            "provider",
            "provider_customer_id",
        ),
    )


class PaySetupIntent(ModelBase):
    __tablename__ = "pay_setup_intents"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    user_id: Mapped[Optional[str]] = mapped_column(user_id_type(), index=True, nullable=True)
    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_setup_intent_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    status: Mapped[str] = mapped_column(
        String(32), index=True, nullable=False
    )  # requires_action|succeeded|canceled|processing
    client_secret: Mapped[Optional[str]] = mapped_column(String(255))
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )

    __table_args__ = (
        user_fk_constraint("user_id", ondelete="SET NULL"),
        Index("ix_pay_setup_intents_tenant_provider", "tenant_id", "provider"),
    )


class PayDispute(ModelBase):
    __tablename__ = "pay_disputes"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_dispute_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    provider_charge_id: Mapped[Optional[str]] = mapped_column(String(128), index=True)
    amount: Mapped[int] = mapped_column(Numeric(18, 0), nullable=False)
    currency: Mapped[str] = mapped_column(String(8), nullable=False)
    reason: Mapped[Optional[str]] = mapped_column(String(64))
    status: Mapped[str] = mapped_column(
        String(32), index=True, nullable=False
    )  # needs_response|under_review|won|lost|warning_closed
    evidence_due_by: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )


class PayPayout(ModelBase):
    __tablename__ = "pay_payouts"

    id: Mapped[str] = mapped_column(String(64), primary_key=True)

    tenant_id: Mapped[str] = mapped_column(String(TENANT_ID_LEN), index=True, nullable=False)

    provider: Mapped[str] = mapped_column(String(32), index=True, nullable=False)
    provider_payout_id: Mapped[str] = mapped_column(
        String(128), unique=True, index=True, nullable=False
    )
    amount: Mapped[int] = mapped_column(Numeric(18, 0), nullable=False)
    currency: Mapped[str] = mapped_column(String(8), nullable=False)
    status: Mapped[str] = mapped_column(
        String(32), index=True, nullable=False
    )  # pending|in_transit|paid|canceled|failed
    arrival_date: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))
    type: Mapped[Optional[str]] = mapped_column(String(32))  # bank_account|card|...
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )



from __future__ import annotations

from typing import Optional

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from .models import (
    LedgerEntry,
    PayCustomer,
    PayDispute,
    PayEvent,
    PayIntent,
    PayInvoice,
    PayPaymentMethod,
    PayPayout,
    PayPrice,
    PayProduct,
    PaySetupIntent,
    PaySubscription,
)
from .provider.registry import get_provider_registry
from .schemas import (
    BalanceSnapshotOut,
    CaptureIn,
    CustomerOut,
    CustomerUpsertIn,
    DisputeOut,
    IntentCreateIn,
    IntentListFilter,
    IntentOut,
    InvoiceCreateIn,
    InvoiceLineItemIn,
    InvoiceOut,
    InvoicesListFilter,
    PaymentMethodAttachIn,
    PaymentMethodOut,
    PayoutOut,
    PriceCreateIn,
    PriceOut,
    ProductCreateIn,
    ProductOut,
    RefundIn,
    SetupIntentOut,
    StatementRow,
    SubscriptionCreateIn,
    SubscriptionOut,
    SubscriptionUpdateIn,
    UsageRecordIn,
)
from .settings import get_payments_settings


def _default_provider_name() -> str:
    return get_payments_settings().default_provider


class PaymentsService:

    def __init__(self, session: AsyncSession, provider_name: Optional[str] = None):
        self.session = session
        self._provider_name = (provider_name or _default_provider_name()).lower()
        self._adapter = None  # resolved on first use

    # --- internal helpers -----------------------------------------------------

    def _get_adapter(self):
        if self._adapter is not None:
            return self._adapter
        reg = get_provider_registry()
        # Try to fetch the named adapter; if missing, raise a helpful error
        try:
            self._adapter = reg.get(self._provider_name)
        except Exception as e:
            raise RuntimeError(
                f"No payments adapter registered for '{self._provider_name}'. "
                "Install and register a provider (e.g., `stripe`) OR pass a custom adapter via "
                "`add_payments(app, adapters=[...])`. If you only need DB endpoints (like "
                "`/payments/transactions`), this error will not occur unless you call a provider API."
            ) from e
        return self._adapter

    # --- internal event dispatcher (shared by webhook + replay) ---------------
    async def _dispatch_event(self, provider: str, parsed: dict) -> None:
        typ = parsed.get("type", "")
        obj = parsed.get("data") or {}

        if provider == "stripe":
            if typ == "payment_intent.succeeded":
                await self._post_sale(obj)
            elif typ == "charge.refunded":
                await self._post_refund(obj)
            elif typ == "charge.captured":
                await self._post_capture(obj)

    # --- Customers ------------------------------------------------------------

    async def ensure_customer(self, data: CustomerUpsertIn) -> CustomerOut:
        adapter = self._get_adapter()
        out = await adapter.ensure_customer(data)
        # upsert local row
        existing = await self.session.scalar(
            select(PayCustomer).where(
                PayCustomer.provider == out.provider,
                PayCustomer.provider_customer_id == out.provider_customer_id,
            )
        )
        if not existing:
            # If your PayCustomer model has additional columns (email/name), include them here.
            self.session.add(
                PayCustomer(
                    provider=out.provider,
                    provider_customer_id=out.provider_customer_id,
                    user_id=data.user_id,
                )
            )
        return out

    # --- Intents --------------------------------------------------------------

    async def create_intent(self, user_id: Optional[str], data: IntentCreateIn) -> IntentOut:
        adapter = self._get_adapter()
        out = await adapter.create_intent(data, user_id=user_id)
        self.session.add(
            PayIntent(
                provider=out.provider,
                provider_intent_id=out.provider_intent_id,
                user_id=user_id,
                amount=out.amount,
                currency=out.currency,
                status=out.status,
                client_secret=out.client_secret,
            )
        )
        return out

    async def confirm_intent(self, provider_intent_id: str) -> IntentOut:
        adapter = self._get_adapter()
        out = await adapter.confirm_intent(provider_intent_id)
        pi = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == provider_intent_id)
        )
        if pi:
            pi.status = out.status
            pi.client_secret = out.client_secret or pi.client_secret
        return out

    async def cancel_intent(self, provider_intent_id: str) -> IntentOut:
        adapter = self._get_adapter()
        out = await adapter.cancel_intent(provider_intent_id)
        pi = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == provider_intent_id)
        )
        if pi:
            pi.status = out.status
        return out

    async def refund(self, provider_intent_id: str, data: RefundIn) -> IntentOut:
        adapter = self._get_adapter()
        out = await adapter.refund(provider_intent_id, data)
        return out

    # --- Webhooks -------------------------------------------------------------

    async def handle_webhook(self, provider: str, signature: str | None, payload: bytes) -> dict:
        adapter = self._get_adapter()
        parsed = await adapter.verify_and_parse_webhook(signature, payload)
        self.session.add(
            PayEvent(
                provider=provider,
                provider_event_id=parsed["id"],
                type=parsed.get("type", ""),
                payload_json=parsed,
            )
        )

        await self._dispatch_event(provider, parsed)
        return {"ok": True}

    # --- Ledger postings ------------------------------------------------------

    async def _post_sale(self, pi_obj: dict):
        provider_intent_id = pi_obj.get("id")
        amount = int(pi_obj.get("amount") or 0)
        currency = str(pi_obj.get("currency") or "USD").upper()
        intent = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == provider_intent_id)
        )
        if intent:
            intent.status = "succeeded"
            self.session.add(
                LedgerEntry(
                    provider=intent.provider,
                    provider_ref=provider_intent_id,
                    user_id=intent.user_id,
                    amount=+amount,
                    currency=currency,
                    kind="payment",
                    status="posted",
                )
            )

    async def _post_capture(self, charge_obj: dict):
        amount = int(charge_obj.get("amount") or 0)
        currency = str(charge_obj.get("currency") or "USD").upper()
        pi_id = charge_obj.get("payment_intent") or ""
        intent = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == pi_id)
        )
        if intent:
            self.session.add(
                LedgerEntry(
                    provider=intent.provider,
                    provider_ref=charge_obj.get("id"),
                    user_id=intent.user_id,
                    amount=+amount,
                    currency=currency,
                    kind="capture",
                    status="posted",
                )
            )

    async def _post_refund(self, charge_obj: dict):
        amount = int(charge_obj.get("amount_refunded") or 0)
        currency = str(charge_obj.get("currency") or "USD").upper()
        pi_id = charge_obj.get("payment_intent") or ""
        intent = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == pi_id)
        )
        if intent and amount > 0:
            self.session.add(
                LedgerEntry(
                    provider=intent.provider,
                    provider_ref=charge_obj.get("id"),
                    user_id=intent.user_id,
                    amount=+amount,
                    currency=currency,
                    kind="refund",
                    status="posted",
                )
            )

    async def attach_payment_method(self, data: PaymentMethodAttachIn) -> PaymentMethodOut:
        out = await self._get_adapter().attach_payment_method(data)
        # Upsert locally for quick listing
        pm = PayPaymentMethod(
            provider=out.provider,
            provider_customer_id=out.provider_customer_id,
            provider_method_id=out.provider_method_id,
            brand=out.brand,
            last4=out.last4,
            exp_month=out.exp_month,
            exp_year=out.exp_year,
            is_default=out.is_default,
        )
        self.session.add(pm)
        return out

    async def list_payment_methods(self, provider_customer_id: str) -> list[PaymentMethodOut]:
        return await self._get_adapter().list_payment_methods(provider_customer_id)

    async def detach_payment_method(self, provider_method_id: str) -> None:
        await self._get_adapter().detach_payment_method(provider_method_id)

    async def set_default_payment_method(
        self, provider_customer_id: str, provider_method_id: str
    ) -> None:
        await self._get_adapter().set_default_payment_method(
            provider_customer_id, provider_method_id
        )

    # --- Products/Prices ---
    async def create_product(self, data: ProductCreateIn) -> ProductOut:
        out = await self._get_adapter().create_product(data)
        self.session.add(
            PayProduct(
                provider=out.provider,
                provider_product_id=out.provider_product_id,
                name=out.name,
                active=out.active,
            )
        )
        return out

    async def create_price(self, data: PriceCreateIn) -> PriceOut:
        out = await self._get_adapter().create_price(data)
        self.session.add(
            PayPrice(
                provider=out.provider,
                provider_price_id=out.provider_price_id,
                provider_product_id=out.provider_product_id,
                currency=out.currency,
                unit_amount=out.unit_amount,
                interval=out.interval,
                trial_days=out.trial_days,
                active=out.active,
            )
        )
        return out

    # --- Subscriptions ---
    async def create_subscription(self, data: SubscriptionCreateIn) -> SubscriptionOut:
        out = await self._get_adapter().create_subscription(data)
        self.session.add(
            PaySubscription(
                provider=out.provider,
                provider_subscription_id=out.provider_subscription_id,
                provider_price_id=out.provider_price_id,
                status=out.status,
                quantity=out.quantity,
                cancel_at_period_end=out.cancel_at_period_end,
            )
        )
        return out

    async def update_subscription(
        self, provider_subscription_id: str, data: SubscriptionUpdateIn
    ) -> SubscriptionOut:
        out = await self._get_adapter().update_subscription(provider_subscription_id, data)
        # Optionally reflect status/quantity locally (query + update if exists)
        return out

    async def cancel_subscription(
        self, provider_subscription_id: str, at_period_end: bool = True
    ) -> SubscriptionOut:
        out = await self._get_adapter().cancel_subscription(provider_subscription_id, at_period_end)
        return out

    # --- Invoices ---
    async def create_invoice(self, data: InvoiceCreateIn) -> InvoiceOut:
        out = await self._get_adapter().create_invoice(data)
        self.session.add(
            PayInvoice(
                provider=out.provider,
                provider_invoice_id=out.provider_invoice_id,
                provider_customer_id=out.provider_customer_id,
                status=out.status,
                amount_due=out.amount_due,
                currency=out.currency,
                hosted_invoice_url=out.hosted_invoice_url,
                pdf_url=out.pdf_url,
            )
        )
        return out

    async def finalize_invoice(self, provider_invoice_id: str) -> InvoiceOut:
        return await self._get_adapter().finalize_invoice(provider_invoice_id)

    async def void_invoice(self, provider_invoice_id: str) -> InvoiceOut:
        return await self._get_adapter().void_invoice(provider_invoice_id)

    async def pay_invoice(self, provider_invoice_id: str) -> InvoiceOut:
        return await self._get_adapter().pay_invoice(provider_invoice_id)

    # --- Intents QoL ---
    async def get_intent(self, provider_intent_id: str) -> IntentOut:
        return await self._get_adapter().hydrate_intent(provider_intent_id)

    # --- Statements/Rollups ---
    async def daily_statements_rollup(
        self, date_from: str | None = None, date_to: str | None = None
    ) -> list[StatementRow]:
        from datetime import datetime

        from sqlalchemy import func

        q = select(
            func.date_trunc("day", LedgerEntry.ts).label("day"),
            LedgerEntry.currency,
            func.sum(func.case((LedgerEntry.kind == "payment", LedgerEntry.amount), else_=0)).label(
                "gross"
            ),
            func.sum(func.case((LedgerEntry.kind == "refund", LedgerEntry.amount), else_=0)).label(
                "refunds"
            ),
            func.sum(func.case((LedgerEntry.kind == "fee", LedgerEntry.amount), else_=0)).label(
                "fees"
            ),
            func.count().label("count"),
        )
        if date_from:
            try:
                q = q.where(LedgerEntry.ts >= datetime.fromisoformat(date_from))
            except Exception:
                pass
        if date_to:
            try:
                q = q.where(LedgerEntry.ts <= datetime.fromisoformat(date_to))
            except Exception:
                pass
        q = q.group_by(func.date_trunc("day", LedgerEntry.ts), LedgerEntry.currency).order_by(
            func.date_trunc("day", LedgerEntry.ts).desc()
        )

        rows = (await self.session.execute(q)).all()
        out: list[StatementRow] = []
        for day, currency, gross, refunds, fees, count in rows:
            gross = int(gross or 0)
            refunds = int(refunds or 0)
            fees = int(fees or 0)
            out.append(
                StatementRow(
                    period_start=day.isoformat(),
                    period_end=day.isoformat(),
                    currency=str(currency).upper(),
                    gross=gross,
                    refunds=refunds,
                    fees=fees,
                    net=gross - refunds - fees,
                    count=int(count or 0),
                )
            )
        return out

    async def capture_intent(self, provider_intent_id: str, data: CaptureIn) -> IntentOut:
        out = await self._get_adapter().capture_intent(
            provider_intent_id, amount=int(data.amount) if data.amount is not None else None
        )
        pi = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == provider_intent_id)
        )
        if pi:
            pi.status = out.status
            if out.status in ("succeeded", "requires_capture"):  # Stripe specifics vary
                pi.captured = True if out.status == "succeeded" else pi.captured
        return out

    async def list_intents(self, f: IntentListFilter) -> tuple[list[IntentOut], str | None]:
        return await self._get_adapter().list_intents(
            customer_provider_id=f.customer_provider_id,
            status=f.status,
            limit=f.limit or 50,
            cursor=f.cursor,
        )

    # ---- Invoices: lines/list/get/preview ----
    async def add_invoice_line_item(
        self, provider_invoice_id: str, data: InvoiceLineItemIn
    ) -> dict:
        return await self._get_adapter().add_invoice_line_item(provider_invoice_id, data)

    async def list_invoices(self, f: InvoicesListFilter) -> tuple[list[InvoiceOut], str | None]:
        return await self._get_adapter().list_invoices(
            customer_provider_id=f.customer_provider_id,
            status=f.status,
            limit=f.limit or 50,
            cursor=f.cursor,
        )

    async def get_invoice(self, provider_invoice_id: str) -> InvoiceOut:
        return await self._get_adapter().get_invoice(provider_invoice_id)

    async def preview_invoice(
        self, customer_provider_id: str, subscription_id: str | None
    ) -> InvoiceOut:
        return await self._get_adapter().preview_invoice(
            customer_provider_id=customer_provider_id, subscription_id=subscription_id
        )

    # ---- Metered usage ----
    async def create_usage_record(self, data: UsageRecordIn) -> dict:
        return await self._get_adapter().create_usage_record(data)

    # --- Setup Intents --------------------------------------------------------

    async def create_setup_intent(self, payment_method_types: list[str]) -> SetupIntentOut:
        out = await self._get_adapter().create_setup_intent(payment_method_types)
        self.session.add(
            PaySetupIntent(
                provider=out.provider,
                provider_setup_intent_id=out.provider_setup_intent_id,
                user_id=None,  # set if you thread a user_id; your router currently doesn't
                status=out.status,
                client_secret=out.client_secret,
            )
        )
        return out

    async def confirm_setup_intent(self, provider_setup_intent_id: str) -> SetupIntentOut:
        out = await self._get_adapter().confirm_setup_intent(provider_setup_intent_id)
        row = await self.session.scalar(
            select(PaySetupIntent).where(
                PaySetupIntent.provider_setup_intent_id == provider_setup_intent_id
            )
        )
        if row:
            row.status = out.status
            row.client_secret = out.client_secret or row.client_secret
        return out

    async def get_setup_intent(self, provider_setup_intent_id: str) -> SetupIntentOut:
        out = await self._get_adapter().get_setup_intent(provider_setup_intent_id)
        # opportunistic upsert
        row = await self.session.scalar(
            select(PaySetupIntent).where(
                PaySetupIntent.provider_setup_intent_id == provider_setup_intent_id
            )
        )
        if row:
            row.status = out.status
            row.client_secret = out.client_secret or row.client_secret
        else:
            self.session.add(
                PaySetupIntent(
                    provider=out.provider,
                    provider_setup_intent_id=out.provider_setup_intent_id,
                    user_id=None,
                    status=out.status,
                    client_secret=out.client_secret,
                )
            )
        return out

    # --- SCA / 3DS resume -----------------------------------------------------
    async def resume_intent_after_action(self, provider_intent_id: str) -> IntentOut:
        out = await self._get_adapter().resume_intent_after_action(provider_intent_id)
        pi = await self.session.scalar(
            select(PayIntent).where(PayIntent.provider_intent_id == provider_intent_id)
        )
        if pi:
            pi.status = out.status
            pi.client_secret = out.client_secret or pi.client_secret
        return out

    # --- Disputes -------------------------------------------------------------
    async def list_disputes(
        self, *, status: Optional[str], limit: int, cursor: Optional[str]
    ) -> tuple[list[DisputeOut], Optional[str]]:
        return await self._get_adapter().list_disputes(status=status, limit=limit, cursor=cursor)

    async def get_dispute(self, provider_dispute_id: str) -> DisputeOut:
        out = await self._get_adapter().get_dispute(provider_dispute_id)
        # Upsert locally
        row = await self.session.scalar(
            select(PayDispute).where(PayDispute.provider_dispute_id == provider_dispute_id)
        )
        if row:
            row.status = out.status
            row.amount = out.amount
            row.currency = out.currency
        else:
            self.session.add(
                PayDispute(
                    provider=out.provider,
                    provider_dispute_id=out.provider_dispute_id,
                    provider_charge_id=None,  # set if adapter returns it
                    amount=out.amount,
                    currency=out.currency,
                    reason=out.reason,
                    status=out.status,
                )
            )
        return out

    async def submit_dispute_evidence(self, provider_dispute_id: str, evidence: dict) -> DisputeOut:
        out = await self._get_adapter().submit_dispute_evidence(provider_dispute_id, evidence)
        # reflect status
        row = await self.session.scalar(
            select(PayDispute).where(PayDispute.provider_dispute_id == provider_dispute_id)
        )
        if row:
            row.status = out.status
        return out

    # --- Balance --------------------------------------------------------------
    async def get_balance_snapshot(self) -> BalanceSnapshotOut:
        return await self._get_adapter().get_balance_snapshot()

    # --- Payouts --------------------------------------------------------------
    async def list_payouts(
        self, *, limit: int, cursor: Optional[str]
    ) -> tuple[list[PayoutOut], Optional[str]]:
        return await self._get_adapter().list_payouts(limit=limit, cursor=cursor)

    async def get_payout(self, provider_payout_id: str) -> PayoutOut:
        out = await self._get_adapter().get_payout(provider_payout_id)
        # Upsert locally
        row = await self.session.scalar(
            select(PayPayout).where(PayPayout.provider_payout_id == provider_payout_id)
        )
        if row:
            row.status = out.status
            row.amount = out.amount
            row.currency = out.currency
            # arrival_date/type optional; update if present
        else:
            self.session.add(
                PayPayout(
                    provider=out.provider,
                    provider_payout_id=out.provider_payout_id,
                    amount=out.amount,
                    currency=out.currency,
                    status=out.status,
                    # arrival_date/out.type if you add them onto PayoutOut
                )
            )
        return out

    # --- Webhook replay -------------------------------------------------------
    async def replay_webhooks(
        self, since: Optional[str], until: Optional[str], event_ids: list[str]
    ) -> int:
        from datetime import datetime

        q = select(PayEvent).where(PayEvent.provider == self._provider_name)
        if event_ids:
            q = q.where(PayEvent.provider_event_id.in_(event_ids))
        else:
            # ISO8601 strings expected; ignore parsing errors safely
            if since:
                try:
                    q = q.where(PayEvent.received_at >= datetime.fromisoformat(since))
                except Exception:
                    pass
            if until:
                try:
                    q = q.where(PayEvent.received_at <= datetime.fromisoformat(until))
                except Exception:
                    pass

        rows = (await self.session.execute(q)).scalars().all()
        for ev in rows:
            await self._dispatch_event(ev.provider, ev.payload_json)

        return len(rows)



from __future__ import annotations

from fastapi import FastAPI

from .mutators import auth_mutator
from .pipeline import apply_mutators


def _normalize_security_list(sec: list | None, *, drop_schemes: set[str] = None) -> list:
    if not sec:
        return []
    drop_schemes = drop_schemes or set()
    cleaned = []
    for item in sec:
        if not isinstance(item, dict):
            continue
        kept = {k: v for k, v in item.items() if k not in drop_schemes}
        if kept:
            cleaned.append(kept)
    # dedupe exact dicts
    seen_dicts = set()
    unique = []
    for item in cleaned:
        canon = tuple(sorted((k, tuple(v or [])) for k, v in item.items()))
        if canon in seen_dicts:
            continue
        seen_dicts.add(canon)
        unique.append(item)
    # dedupe single-scheme repeats
    seen_schemes = set()
    final = []
    for item in unique:
        if len(item) == 1:
            scheme = next(iter(item))
            if scheme in seen_schemes:
                continue
            seen_schemes.add(scheme)
        final.append(item)
    return final


def install_openapi_auth(app: FastAPI, *, include_api_key: bool = False) -> None:
    apply_mutators(app, auth_mutator(include_api_key))



from fastapi import HTTPException, Request, status


def require_if_match(request: Request, current_etag: str):
    val = request.headers.get("If-Match")
    if not val:
        raise HTTPException(
            status_code=status.HTTP_428_PRECONDITION_REQUIRED,
            detail="If-Match header required for update.",
        )
    if current_etag not in [t.strip() for t in val.split(",")]:
        raise HTTPException(
            status_code=status.HTTP_412_PRECONDITION_FAILED, detail="ETag precondition failed."
        )



from __future__ import annotations

from datetime import datetime, timezone
from typing import Annotated, Any, Callable, Optional

from fastapi import Depends, HTTPException, Request
from fastapi.security import APIKeyCookie, APIKeyHeader, OAuth2PasswordBearer
from sqlalchemy import select

from svc_infra.api.fastapi.auth.settings import get_auth_settings
from svc_infra.api.fastapi.auth.state import get_auth_state, get_user_scope_resolver
from svc_infra.api.fastapi.db.sql.session import SqlSessionDep
from svc_infra.api.fastapi.paths.prefix import USER_PREFIX
from svc_infra.api.fastapi.paths.user import LOGIN_PATH
from svc_infra.db.sql.apikey import get_apikey_model

# ---------- OpenAPI security schemes (appear in docs) ----------
auth_login_path = USER_PREFIX + LOGIN_PATH
oauth2_scheme_optional = OAuth2PasswordBearer(tokenUrl=auth_login_path, auto_error=False)
cookie_auth_optional = APIKeyCookie(name=get_auth_settings().auth_cookie_name, auto_error=False)
api_key_header = APIKeyHeader(name="X-API-Key", auto_error=False)


# ---------- Principal ----------
class Principal:
    """Unified identity: user via JWT/cookie or service via API key."""

    def __init__(
        self, *, user=None, scopes: list[str] | None = None, via: str = "jwt", api_key=None
    ):
        self.user = user
        self.scopes = scopes or []
        self.via = via  # "jwt" | "cookie" | "api_key"
        self.api_key = api_key


# ---------- Resolvers ----------
async def resolve_api_key(
    request: Request,
    session: SqlSessionDep,
) -> Optional[Principal]:
    raw = (request.headers.get("x-api-key") or "").strip()
    if not raw:
        return None
    ApiKey = get_apikey_model()
    prefix = ""
    parts = raw.split("_", 2)
    if len(parts) >= 3 and parts[0] == "ak":
        prefix = parts[1][:12]

    apikey = None
    if prefix:
        apikey = (
            (await session.execute(select(ApiKey).where(ApiKey.key_prefix == prefix)))
            .scalars()
            .first()
        )
    if not apikey:
        raise HTTPException(401, "invalid_api_key")

    from hmac import compare_digest

    if not compare_digest(ApiKey.hash(raw), apikey.key_hash):
        raise HTTPException(401, "invalid_api_key")
    if not apikey.active:
        raise HTTPException(401, "api_key_revoked")
    if apikey.expires_at and datetime.now(timezone.utc) > apikey.expires_at:
        raise HTTPException(401, "api_key_expired")

    apikey.mark_used()
    await session.flush()
    return Principal(user=apikey.user, scopes=apikey.scopes, via="api_key", api_key=apikey)


async def resolve_bearer_or_cookie_principal(
    request: Request, session: SqlSessionDep
) -> Optional[Principal]:
    st = get_auth_settings()
    raw_auth = (request.headers.get("authorization") or "").strip()
    token = raw_auth.split(" ", 1)[1].strip() if raw_auth.lower().startswith("bearer ") else ""
    if not token:
        token = (request.cookies.get(st.auth_cookie_name) or "").strip()
    if not token:
        return None

    UserModel, get_strategy, _ = get_auth_state()
    strategy = get_strategy()

    from fastapi_users.manager import BaseUserManager, UUIDIDMixin
    from fastapi_users_db_sqlalchemy import SQLAlchemyUserDatabase

    user_db = SQLAlchemyUserDatabase(session, UserModel)

    class _ShimManager(UUIDIDMixin, BaseUserManager[Any, Any]):
        reset_password_token_secret = "unused"
        verification_token_secret = "unused"

        def __init__(self, db):
            super().__init__(db)

    user_manager = _ShimManager(user_db)

    try:
        user = await strategy.read_token(token, user_manager)
    except Exception:
        return None
    if not user:
        return None

    db_user = await session.get(UserModel, user.id)
    if not db_user:
        return None
    if not getattr(db_user, "is_active", True):
        raise HTTPException(401, "account_disabled")

    via = "jwt" if raw_auth else "cookie"
    user_scopes = get_user_scope_resolver()(db_user)
    # dedupe while keeping order
    scopes = list(dict.fromkeys(user_scopes))
    return Principal(user=db_user, scopes=scopes, via=via)


async def _current_principal(
    request: Request,
    session: SqlSessionDep,
    jwt_or_cookie: Optional[Principal] = Depends(resolve_bearer_or_cookie_principal),
    ak: Optional[Principal] = Depends(resolve_api_key),
) -> Principal:
    if jwt_or_cookie:
        return jwt_or_cookie
    if ak:
        return ak
    raise HTTPException(401, "Missing credentials")


async def _optional_principal(
    request: Request,
    session: SqlSessionDep,
    jwt_or_cookie: Optional[Principal] = Depends(resolve_bearer_or_cookie_principal),
    ak: Optional[Principal] = Depends(resolve_api_key),
) -> Optional[Principal]:
    return jwt_or_cookie or ak or None


# ---------- DX: types for endpoint params ----------
Identity = Annotated[Principal, Depends(_current_principal)]
OptionalIdentity = Annotated[Principal | None, Depends(_optional_principal)]

# ---------- DX: constants for router-level dependencies ----------
RequireIdentity = Depends(_current_principal)  # use inside router dependencies=[...]
AllowIdentity = Depends(_optional_principal)  # same, but optional


# ---------- DX: small guard factories ----------
def RequireRoles(*roles: str, resolver: Callable[[Any], list[str]] | None = None):
    async def _guard(p: Identity):
        have = set((resolver(p.user) if resolver else getattr(p.user, "roles", []) or []))
        if not set(roles).issubset(have):
            raise HTTPException(403, "forbidden")
        return p

    return Depends(_guard)


def RequireScopes(*needed: str):
    async def _guard(p: Identity):  # Identity = resolves to Principal
        if not set(needed).issubset(set(p.scopes or [])):
            raise HTTPException(403, "insufficient_scope")
        return p

    return Depends(_guard)


def RequireAnyScope(*candidates: str):
    async def _guard(p: Identity):
        if not set(p.scopes or []) & set(candidates):
            raise HTTPException(403, "insufficient_scope")
        return p

    return Depends(_guard)


def RequireUser():
    async def _guard(p: Identity):
        if not p.user:
            raise HTTPException(401, "user_required")
        return p

    return Depends(_guard)


def RequireService():
    async def _guard(p: Identity):
        if not p.api_key:
            raise HTTPException(401, "api_key_required")
        return p

    return Depends(_guard)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel

# --- Email OTP store (replace with Redis in prod) ---
EMAIL_OTP_STORE: dict[str, dict] = {}  # key = uid (or jti), value={hash,exp,attempts,next_send}


class StartSetupOut(BaseModel):
    otpauth_url: str
    secret: str
    qr_svg: str | None = None  # optional: inline SVG


class ConfirmSetupIn(BaseModel):
    code: str


class VerifyMFAIn(BaseModel):
    code: str
    pre_token: str


class DisableMFAIn(BaseModel):
    code: str | None = None
    recovery_code: str | None = None


class RecoveryCodesOut(BaseModel):
    codes: list[str]


class SendEmailCodeIn(BaseModel):
    pre_token: str


class SendEmailCodeOut(BaseModel):
    sent: bool = True
    cooldown_seconds: int = 60


class MFAStatusOut(BaseModel):
    enabled: bool
    methods: list[str]
    confirmed_at: datetime | None = None
    email_mask: str | None = None
    email_otp: dict | None = None


class MFAProof(BaseModel):
    code: str | None = None
    pre_token: str | None = None


class DisableAccountIn(BaseModel):
    reason: Optional[str] = None
    mfa: Optional[MFAProof] = None


class DeleteAccountIn(BaseModel):
    mfa: Optional[MFAProof] = None



from __future__ import annotations

from typing import Dict, Optional

from ..settings import get_payments_settings
from .base import ProviderAdapter


class ProviderRegistry:
    def __init__(self):
        self._adapters: Dict[str, ProviderAdapter] = {}

    def register(self, adapter: ProviderAdapter):
        self._adapters[adapter.name] = adapter

    def get(self, name: Optional[str] = None) -> ProviderAdapter:
        settings = get_payments_settings()
        key = (name or settings.default_provider).lower()
        if key not in self._adapters:
            raise RuntimeError(f"No payments adapter registered for '{key}'")
        return self._adapters[key]


_REGISTRY: Optional[ProviderRegistry] = None


def get_provider_registry() -> ProviderRegistry:
    global _REGISTRY
    if _REGISTRY is None:
        _REGISTRY = ProviderRegistry()
    return _REGISTRY



from __future__ import annotations

from pydantic import BaseModel, field_validator, model_validator


class Contact(BaseModel):
    name: str | None = None
    email: str | None = None
    url: str | None = None


class License(BaseModel):
    name: str | None = None
    url: str | None = None


class ServiceInfo(BaseModel):
    name: str = "Service Infrastructure App"
    release: str = "0.1.0"
    description: str | None = None
    terms_of_service: str | None = None
    contact: Contact | None = None
    license: License | None = None


class VersionInfo(BaseModel):
    # All optional; only fields you set will override the base ServiceInfo.
    title: str | None = None  # overrides info.title
    version_label: str | None = None  # overrides info.version (label only)
    description: str | None = None
    terms_of_service: str | None = None
    contact: Contact | None = None
    license: License | None = None


class APIVersionSpec(BaseModel):
    tag: str | int = "v0"
    routers_package: str | None = None
    public_base_url: str | None = None  # None -> relative "/vN"
    docs: bool | None = None
    include_api_key: bool | None = None  # per-version auth toggle

    # NEW: single, cohesive override object
    info: VersionInfo | None = None

    # ---- Back-compat shim (optional): map old per-field overrides into info ----
    # If you want to keep supporting the old fields for a while:
    description: str | None = None
    terms_of_service: str | None = None
    contact: Contact | None = None
    license: License | None = None

    @model_validator(mode="after")
    def _back_compat_into_info(self):
        if any([self.description, self.terms_of_service, self.contact, self.license]):
            vi = self.info or VersionInfo()
            if self.description is not None:
                vi.description = self.description
            if self.terms_of_service is not None:
                vi.terms_of_service = self.terms_of_service
            if self.contact is not None:
                vi.contact = self.contact
            if self.license is not None:
                vi.license = self.license
            object.__setattr__(self, "info", vi)
        return self

    @field_validator("tag", mode="before")
    @classmethod
    def _coerce_tag(cls, v):
        if isinstance(v, int):
            return f"v{v}"
        s = str(v or "").strip().lstrip("/")
        if not s:
            return "v0"
        if s.startswith("v"):
            return s
        if s.isdigit():
            return f"v{s}"
        return s



from __future__ import annotations

import hashlib
import hmac
import os
import uuid
from datetime import datetime, timezone
from typing import Optional, Type

from sqlalchemy import JSON, Boolean, DateTime, ForeignKey, Index, String, UniqueConstraint, text
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import Mapped, declared_attr, mapped_column, relationship

from svc_infra.db.sql.base import ModelBase
from svc_infra.db.sql.types import GUID

_APIKEY_HMAC_SECRET = os.getenv("APIKEY_HASH_SECRET") or "change-me-low-entropy-dev"


def _hmac_sha256(s: str) -> str:
    return hmac.new(_APIKEY_HMAC_SECRET.encode(), s.encode(), hashlib.sha256).hexdigest()


def _now() -> datetime:
    return datetime.now(timezone.utc)


# -------------------- Factory & registry --------------------

_ApiKeyModel: Optional[type] = None


def get_apikey_model() -> type:
    """Return the bound ApiKey model (or raise if not enabled)."""
    if _ApiKeyModel is None:
        raise RuntimeError("ApiKey model is not enabled. Call bind_apikey_model(...) first.")
    return _ApiKeyModel


def bind_apikey_model(user_model: Type, *, table_name: str = "api_keys") -> type:
    """
    Create and register an ApiKey model bound to the provided user_model and table name.
    Call this once during app boot (e.g., inside add_auth_users when enable_api_keys=True).
    """

    class ApiKey(ModelBase):  # type: ignore[misc, valid-type]
        __tablename__ = table_name

        id: Mapped[uuid.UUID] = mapped_column(GUID(), primary_key=True, default=uuid.uuid4)

        @declared_attr
        def user_id(cls) -> Mapped[uuid.UUID | None]:  # noqa: N805
            return mapped_column(
                GUID(),
                ForeignKey(f"{user_model.__tablename__}.id", ondelete="SET NULL"),
                nullable=True,
                index=True,
            )

        @declared_attr
        def user(cls):  # noqa: N805
            return relationship(user_model.__name__, lazy="selectin")

        name: Mapped[str] = mapped_column(String(128), nullable=False)

        # hash + short prefix for lookup
        key_prefix: Mapped[str] = mapped_column(String(12), index=True, nullable=False)
        key_hash: Mapped[str] = mapped_column(String(64), nullable=False)  # hex sha256

        scopes: Mapped[list[str]] = mapped_column(MutableList.as_mutable(JSON), default=list)
        active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
        expires_at: Mapped[datetime | None] = mapped_column(DateTime(timezone=True))
        last_used_at: Mapped[datetime | None] = mapped_column(DateTime(timezone=True))
        meta: Mapped[dict] = mapped_column(MutableDict.as_mutable(JSON), default=dict)

        created_at = mapped_column(
            DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
        )
        updated_at = mapped_column(
            DateTime(timezone=True),
            server_default=text("CURRENT_TIMESTAMP"),
            onupdate=text("CURRENT_TIMESTAMP"),
            nullable=False,
        )

        __table_args__ = (
            UniqueConstraint("key_prefix", name="uq_apikey_prefix"),
            Index("ix_api_keys_user_id", "user_id"),
        )

        # Helpers
        @staticmethod
        def make_secret() -> tuple[str, str, str]:
            """
            Returns (plaintext, prefix, hash). plaintext is shown ONCE to the caller.
            Format: ak_<prefix>_<random>
            """
            import base64
            import secrets

            prefix = secrets.token_urlsafe(6).replace("-", "").replace("_", "")[:8]
            rand = base64.urlsafe_b64encode(secrets.token_bytes(24)).decode().rstrip("=")
            plaintext = f"ak_{prefix}_{rand}"
            return plaintext, prefix, _hmac_sha256(plaintext)

        @staticmethod
        def hash(plaintext: str) -> str:
            return _hmac_sha256(plaintext)

        def mark_used(self):
            self.last_used_at = _now()

    global _ApiKeyModel
    _ApiKeyModel = ApiKey
    return ApiKey


def try_autobind_apikey_model(*, require_env: bool = False) -> Optional[type]:
    """
    If API keys aren’t bound yet, try to discover the User model and bind.
    - If require_env=True, only bind when AUTH_ENABLE_API_KEYS is truthy.
    """
    global _ApiKeyModel
    if _ApiKeyModel is not None:
        return _ApiKeyModel

    if require_env:
        flag = os.getenv("AUTH_ENABLE_API_KEYS", "")
        if str(flag).strip().lower() not in {"1", "true", "yes"}:
            return None

    try:
        from svc_infra.db.sql.base import ModelBase

        # SQLAlchemy 2.x: iterate registry mappers to get mapped classes
        for mapper in list(getattr(ModelBase, "registry").mappers):
            cls = mapper.class_
            if getattr(cls, "__svc_infra_auth_user__", False):
                return bind_apikey_model(cls)  # binds and returns ApiKey
    except Exception:
        return None

    return None



from __future__ import annotations

from typing import Any, Literal, Optional

from pydantic import BaseModel, Field, conint, constr

Currency = constr(pattern=r"^[A-Z]{3}$")
AmountMinor = conint(ge=0)  # minor units (cents)


class CustomerUpsertIn(BaseModel):
    user_id: Optional[str] = None
    email: Optional[str] = None
    name: Optional[str] = None


class CustomerOut(BaseModel):
    id: str
    provider: str
    provider_customer_id: str
    email: Optional[str] = None
    name: Optional[str] = None


class IntentCreateIn(BaseModel):
    amount: AmountMinor = Field(..., description="Minor units (e.g., cents)")
    currency: Currency = Field(..., example="USD")
    description: Optional[str] = None
    capture_method: Literal["automatic", "manual"] = "automatic"
    payment_method_types: list[str] = Field(default_factory=list)  # let provider default


class NextAction(BaseModel):
    type: Optional[str] = None
    data: Optional[dict[str, Any]] = None


class IntentOut(BaseModel):
    id: str
    provider: str
    provider_intent_id: str
    status: str
    amount: AmountMinor
    currency: Currency
    client_secret: Optional[str] = None
    next_action: Optional[NextAction] = None


class RefundIn(BaseModel):
    amount: Optional[AmountMinor] = None
    reason: Optional[str] = None


class TransactionRow(BaseModel):
    id: str
    ts: str
    type: Literal["payment", "refund", "fee", "payout", "capture"]
    amount: int
    currency: Currency
    status: str
    provider: str
    provider_ref: str
    user_id: Optional[str] = None
    net: Optional[int] = None
    fee: Optional[int] = None


class StatementRow(BaseModel):
    period_start: str
    period_end: str
    currency: Currency
    gross: int
    refunds: int
    fees: int
    net: int
    count: int


class PaymentMethodAttachIn(BaseModel):
    customer_provider_id: str
    payment_method_token: str  # provider token (e.g., stripe pm_ or payment_method id)
    make_default: bool = True


class PaymentMethodOut(BaseModel):
    id: str
    provider: str
    provider_customer_id: str
    provider_method_id: str
    brand: Optional[str] = None
    last4: Optional[str] = None
    exp_month: Optional[int] = None
    exp_year: Optional[int] = None
    is_default: bool = False


class ProductCreateIn(BaseModel):
    name: str
    active: bool = True


class ProductOut(BaseModel):
    id: str
    provider: str
    provider_product_id: str
    name: str
    active: bool


class PriceCreateIn(BaseModel):
    provider_product_id: str
    currency: Currency
    unit_amount: AmountMinor
    interval: Optional[Literal["day", "week", "month", "year"]] = None
    trial_days: Optional[int] = None
    active: bool = True


class PriceOut(BaseModel):
    id: str
    provider: str
    provider_price_id: str
    provider_product_id: str
    currency: Currency
    unit_amount: AmountMinor
    interval: Optional[str] = None
    trial_days: Optional[int] = None
    active: bool = True


class SubscriptionCreateIn(BaseModel):
    customer_provider_id: str
    price_provider_id: str
    quantity: int = 1
    trial_days: Optional[int] = None
    proration_behavior: Literal["create_prorations", "none", "always_invoice"] = "create_prorations"


class SubscriptionUpdateIn(BaseModel):
    price_provider_id: Optional[str] = None
    quantity: Optional[int] = None
    cancel_at_period_end: Optional[bool] = None
    proration_behavior: Literal["create_prorations", "none", "always_invoice"] = "create_prorations"


class SubscriptionOut(BaseModel):
    id: str
    provider: str
    provider_subscription_id: str
    provider_price_id: str
    status: str
    quantity: int
    cancel_at_period_end: bool
    current_period_end: Optional[str] = None


class InvoiceCreateIn(BaseModel):
    customer_provider_id: str
    auto_advance: bool = True


class InvoiceOut(BaseModel):
    id: str
    provider: str
    provider_invoice_id: str
    provider_customer_id: str
    status: str
    amount_due: AmountMinor
    currency: Currency
    hosted_invoice_url: Optional[str] = None
    pdf_url: Optional[str] = None


class CaptureIn(BaseModel):
    amount: Optional[AmountMinor] = None  # partial capture supported


class IntentListFilter(BaseModel):
    customer_provider_id: Optional[str] = None
    status: Optional[str] = None
    limit: Optional[int] = Field(default=50, ge=1, le=200)
    cursor: Optional[str] = None  # opaque provider cursor when supported


class UsageRecordIn(BaseModel):
    # Stripe: subscription_item is the target for metered billing.
    # If provider doesn't use subscription_item, allow provider_price_id as fallback.
    subscription_item: Optional[str] = None
    provider_price_id: Optional[str] = None
    quantity: conint(ge=0)
    timestamp: Optional[int] = None  # Unix seconds; provider defaults to "now"
    action: Optional[Literal["increment", "set"]] = "increment"


class InvoiceLineItemIn(BaseModel):
    customer_provider_id: str
    description: Optional[str] = None
    unit_amount: AmountMinor
    currency: Currency
    quantity: Optional[int] = 1
    provider_price_id: Optional[str] = None  # if linked to a price, unit_amount may be ignored


class InvoicesListFilter(BaseModel):
    customer_provider_id: Optional[str] = None
    status: Optional[str] = None
    limit: Optional[int] = Field(default=50, ge=1, le=200)
    cursor: Optional[str] = None


class SetupIntentOut(BaseModel):
    id: str
    provider: str
    provider_setup_intent_id: str
    status: str
    client_secret: Optional[str] = None
    next_action: Optional[NextAction] = None


class DisputeOut(BaseModel):
    id: str
    provider: str
    provider_dispute_id: str
    amount: AmountMinor
    currency: Currency
    reason: Optional[str] = None
    status: str
    evidence_due_by: Optional[str] = None
    created_at: Optional[str] = None


class PayoutOut(BaseModel):
    id: str
    provider: str
    provider_payout_id: str
    amount: AmountMinor
    currency: Currency
    status: str
    arrival_date: Optional[str] = None
    type: Optional[str] = None


class BalanceAmount(BaseModel):
    currency: Currency
    amount: AmountMinor


class BalanceSnapshotOut(BaseModel):
    available: list[BalanceAmount] = Field(default_factory=list)
    pending: list[BalanceAmount] = Field(default_factory=list)



from __future__ import annotations

from typing import Any, Callable, Type

from fastapi import APIRouter
from fastapi.params import Depends
from pydantic import BaseModel

from ..pagination import Paginated, make_pagination_injector
from .utils import _alt_with_slash, _norm_primary


class DualAPIRouter(APIRouter):
    """
    Registers two routes per endpoint:
      • primary: shown in OpenAPI (no trailing slash)
      • alternate: hidden in OpenAPI (with trailing slash)
    Keeps redirect_slashes=False behavior (no 307s).
    """

    def __init__(self, *args, redirect_slashes: bool = False, **kwargs) -> None:
        # Force no implicit 307s; we explicitly add a twin instead.
        super().__init__(*args, redirect_slashes=redirect_slashes, **kwargs)

    # ---------- core helper ----------

    def _dual_decorator(
        self, path: str, methods: list[str], *, show_in_schema: bool = True, **kwargs
    ):
        is_rootish = path in {"", "/"}
        primary = _norm_primary(path or "")
        alt = _alt_with_slash(path or "")

        safe_methods = {"GET", "HEAD", "OPTIONS"}

        def decorator(func):
            if is_rootish:
                # primary root
                self.add_api_route(
                    "", func, methods=methods, include_in_schema=show_in_schema, **kwargs
                )
                # only add the "/" twin for *safe* methods
                if set(m.upper() for m in methods) <= safe_methods:
                    self.add_api_route(
                        "/", func, methods=methods, include_in_schema=False, **kwargs
                    )
                return func

            # non-root unchanged
            self.add_api_route(
                primary, func, methods=methods, include_in_schema=show_in_schema, **kwargs
            )
            if alt != primary:
                self.add_api_route(alt, func, methods=methods, include_in_schema=False, **kwargs)
            return func

        return decorator

    def add_api_route(self, path, endpoint, **kwargs):
        methods = set((kwargs.get("methods") or []))
        for r in self.routes:
            if getattr(r, "path", None) == path and methods & (
                getattr(r, "methods", set()) or set()
            ):
                raise RuntimeError(f"Duplicate route in router: {methods} {path}")
        return super().add_api_route(path, endpoint, **kwargs)

    # ---------- HTTP method shorthands ----------

    def get(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["GET"], show_in_schema=show_in_schema, **kwargs)

    def post(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["POST"], show_in_schema=show_in_schema, **kwargs)

    def patch(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["PATCH"], show_in_schema=show_in_schema, **kwargs)

    def delete(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["DELETE"], show_in_schema=show_in_schema, **kwargs)

    def put(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["PUT"], show_in_schema=show_in_schema, **kwargs)

    def options(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["OPTIONS"], show_in_schema=show_in_schema, **kwargs)

    def head(self, path: str, *_, show_in_schema: bool = True, **kwargs: Any):
        return self._dual_decorator(path, ["HEAD"], show_in_schema=show_in_schema, **kwargs)

    def list(
        self,
        path: str,
        *,
        model: Type[BaseModel],
        envelope: bool = False,
        cursor: bool = True,
        page: bool = True,
        default_limit: int = 50,
        max_limit: int = 200,
        show_in_schema: bool = True,
        **kwargs: Any,
    ):
        """
        Sugar for list endpoints.

        - Auto-inject pagination/filter context (no Depends in your signature).
        - Auto-picks response_model: list[model] or Paginated[model].
        - Works with your OpenAPI mutators which already attach the shared params.
        - Per-route opt-out of OpenAPI param auto-attach: openapi_extra={"x_no_auto_pagination": True}
        """
        # pick response model
        if envelope:
            response_model = Paginated[model]  # type: ignore[index]
        else:
            response_model = list[model]  # type: ignore[index]

        injector = make_pagination_injector(
            envelope=envelope,
            allow_cursor=cursor,
            allow_page=page,
            default_limit=default_limit,
            max_limit=max_limit,
        )

        # ensure our injector runs; don't mutate caller's dependencies
        deps = list(kwargs.get("dependencies") or [])
        deps.append(Depends(injector))
        kwargs["dependencies"] = deps
        kwargs["response_model"] = kwargs.get("response_model") or response_model

        # we still want the dual-registration behavior
        return self._dual_decorator(path, ["GET"], show_in_schema=show_in_schema, **kwargs)

    # ---------- WebSocket ----------

    def websocket(self, path: str, *_, **kwargs: Any):
        """
        Dual-registrations for WebSockets. Starlette doesn't expose OpenAPI for WS,
        so there is no schema visibility knob here.
        """
        primary = _norm_primary(path or "")
        alt = _alt_with_slash(path or "")

        def decorator(func: Callable[..., Any]):
            # Signature must accept (websocket: WebSocket, ...)
            if "dependencies" in kwargs:
                # FastAPI's add_api_websocket_route also accepts dependencies, tags, name, etc.
                self.add_api_websocket_route(primary, func, **kwargs)
                if alt != primary:
                    self.add_api_websocket_route(alt, func, **kwargs)
            else:
                self.add_api_websocket_route(primary, func, **kwargs)
                if alt != primary:
                    self.add_api_websocket_route(alt, func, **kwargs)
            return func

        return decorator



import base64
import hashlib
import os

import pyotp


def _qr_svg_from_uri(uri: str) -> str:
    # Placeholder SVG; most frontends will render their own QR
    return (
        "<svg xmlns='http://www.w3.org/2000/svg' width='280' height='280'>"
        "<rect width='100%' height='100%' fill='#fff'/>"
        f"<text x='10' y='20' font-size='10'>{uri}</text></svg>"
    )


def _random_base32() -> str:
    return pyotp.random_base32(length=32)


def _gen_recovery_codes(n: int, length: int) -> list[str]:
    out = []
    for _ in range(n):
        raw = base64.urlsafe_b64encode(os.urandom(24)).decode().rstrip("=")
        out.append(raw[:length])
    return out


def _gen_numeric_code(n: int = 6) -> str:
    import random

    return "".join(str(random.randrange(10)) for _ in range(n))


def _hash(s: str) -> str:
    return hashlib.sha256(s.encode()).hexdigest()


def _now_utc_ts() -> int:
    from datetime import datetime, timezone

    return int(datetime.now(timezone.utc).timestamp())



import hashlib
import time
from typing import Annotated

from fastapi import Header, HTTPException, Request
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response


class IdempotencyMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, ttl_seconds: int = 24 * 3600, store=None):
        super().__init__(app)
        self.ttl = ttl_seconds
        self.store = store or {}  # replace with Redis

    def _cache_key(self, request, idkey: str):
        body = getattr(request, "_body", None)
        if body is None:
            body = b""

            async def _read():
                data = await request.body()
                request._body = data  # stash for downstream
                return data

            # read once
            # note: starlette Request is awaitable; we read in dispatch below

        sig = hashlib.sha256(
            (
                request.method + "|" + request.url.path + "|" + idkey + "|" + (request._body or b"")
            ).encode()
            if isinstance(request._body, str)
            else (request.method + "|" + request.url.path + "|" + idkey).encode()
            + (request._body or b"")
        ).hexdigest()
        return f"idmp:{sig}"

    async def dispatch(self, request, call_next):
        if request.method in {"POST", "PATCH", "DELETE"}:
            # read & buffer body once
            body = await request.body()
            request._body = body
            idkey = request.headers.get("Idempotency-Key")
            if idkey:
                k = self._cache_key(request, idkey)
                entry = self.store.get(k)
                now = time.time()
                if entry and entry["exp"] > now:
                    cached = entry["resp"]
                    return Response(
                        content=cached["body"],
                        status_code=cached["status"],
                        headers=cached["headers"],
                        media_type=cached.get("media_type"),
                    )
                resp = await call_next(request)
                # cache only 2xx/201 responses
                if 200 <= resp.status_code < 300:
                    body_bytes = b"".join([section async for section in resp.body_iterator])
                    headers = dict(resp.headers)
                    self.store[k] = {
                        "resp": {
                            "status": resp.status_code,
                            "body": body_bytes,
                            "headers": headers,
                            "media_type": resp.media_type,
                        },
                        "exp": now + self.ttl,
                    }
                    return Response(
                        content=body_bytes,
                        status_code=resp.status_code,
                        headers=headers,
                        media_type=resp.media_type,
                    )
                return resp
        return await call_next(request)


async def require_idempotency_key(
    idempotency_key: Annotated[str, Header(alias="Idempotency-Key")],
    request: Request,
) -> None:
    if not idempotency_key.strip():
        raise HTTPException(status_code=400, detail="Idempotency-Key must not be empty.")



import logging
import traceback
from typing import Any, Dict, Optional

from fastapi import Request
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.responses import JSONResponse, Response
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from starlette.exceptions import HTTPException as StarletteHTTPException

from svc_infra.api.fastapi.middleware.errors.exceptions import FastApiException
from svc_infra.app.env import IS_PROD

logger = logging.getLogger(__name__)

PROBLEM_MT = "application/problem+json"


def _trace_id_from_request(request: Request) -> Optional[str]:
    # Try common headers first; fall back to None
    for h in ("x-request-id", "x-correlation-id", "x-trace-id"):
        v = request.headers.get(h)
        if v:
            return v
    return None


def _json_safe(obj):
    if isinstance(obj, dict):
        return {k: _json_safe(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_json_safe(v) for v in obj]
    # Convert any non-primitive (e.g., Exception) to string
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj
    return str(obj)


def problem_response(
    *,
    status: int,
    title: str,
    detail: str | None = None,
    type_uri: str = "about:blank",
    instance: str | None = None,
    code: str | None = None,
    errors: list[dict] | None = None,
    trace_id: str | None = None,
) -> Response:
    body: Dict[str, Any] = {
        "type": type_uri,
        "title": title,
        "status": status,
    }
    if detail is not None:
        body["detail"] = detail
    if instance is not None:
        body["instance"] = instance
    if code is not None:
        body["code"] = code
    if errors:
        body["errors"] = errors
    if trace_id:
        body["trace_id"] = trace_id
    return JSONResponse(status_code=status, content=body, media_type=PROBLEM_MT)


def register_error_handlers(app):
    @app.exception_handler(FastApiException)
    async def handle_app_exception(request: Request, exc: FastApiException):
        trace_id = _trace_id_from_request(request)
        title = exc.title or "Bad Request"
        # In prod, keep 500 messages generic
        detail = (
            exc.detail
            if (not IS_PROD or exc.status_code < 500)
            else "Something went wrong. Please contact support."
        )
        return problem_response(
            status=exc.status_code,
            title=title,
            detail=detail,
            code=exc.code,
            instance=str(request.url),
            trace_id=trace_id,
        )

    @app.exception_handler(RequestValidationError)
    async def handle_validation_error(request: Request, exc: RequestValidationError):
        trace_id = _trace_id_from_request(request)
        raw = exc.errors()
        safe_errors = _json_safe(raw)
        detail = None if IS_PROD else "Validation failed."
        return problem_response(
            status=422,
            title="Unprocessable Entity",
            detail=detail,
            errors=safe_errors if not IS_PROD else None,
            code="VALIDATION_ERROR",
            instance=str(request.url),
            trace_id=trace_id,
        )

    @app.exception_handler(HTTPException)
    async def handle_http_exception(request: Request, exc: HTTPException):
        trace_id = _trace_id_from_request(request)
        title = {401: "Unauthorized", 403: "Forbidden", 404: "Not Found"}.get(
            exc.status_code, "Error"
        )
        detail = (
            exc.detail
            if not IS_PROD or exc.status_code < 500
            else "Something went wrong. Please contact support."
        )
        return problem_response(
            status=exc.status_code,
            title=title,
            detail=str(detail) if isinstance(detail, (dict, list)) else detail,
            code=title.replace(" ", "_").upper(),
            instance=str(request.url),
            trace_id=trace_id,
        )

    @app.exception_handler(StarletteHTTPException)
    async def handle_starlette_http_exception(request: Request, exc: StarletteHTTPException):
        trace_id = _trace_id_from_request(request)
        title = {401: "Unauthorized", 403: "Forbidden", 404: "Not Found"}.get(
            exc.status_code, "Error"
        )
        detail = (
            exc.detail
            if not IS_PROD or exc.status_code < 500
            else "Something went wrong. Please contact support."
        )
        return problem_response(
            status=exc.status_code,
            title=title,
            detail=str(detail) if isinstance(detail, (dict, list)) else detail,
            code=title.replace(" ", "_").upper(),
            instance=str(request.url),
            trace_id=trace_id,
        )

    @app.exception_handler(IntegrityError)
    async def handle_integrity_error(request: Request, exc: IntegrityError):
        trace_id = _trace_id_from_request(request)
        msg = str(getattr(exc, "orig", exc))
        if "duplicate key value" in msg or "UniqueViolation" in msg:
            return problem_response(
                status=409,
                title="Conflict",
                detail="Record already exists.",
                code="CONFLICT",
                instance=str(request.url),
                trace_id=trace_id,
            )
        if "not-null" in msg or "NotNullViolation" in msg:
            return problem_response(
                status=400,
                title="Bad Request",
                detail="Missing required field.",
                code="BAD_REQUEST",
                instance=str(request.url),
                trace_id=trace_id,
            )
        return problem_response(
            status=500,
            title="Internal Server Error",
            detail="Please try again later." if IS_PROD else str(exc),
            code="INTERNAL_ERROR",
            instance=str(request.url),
            trace_id=trace_id,
        )

    @app.exception_handler(SQLAlchemyError)
    async def handle_sqlalchemy_error(request: Request, exc: SQLAlchemyError):
        trace_id = _trace_id_from_request(request)
        return problem_response(
            status=500,
            title="Internal Server Error",
            detail="Please try again later." if IS_PROD else str(exc),
            code="INTERNAL_ERROR",
            instance=str(request.url),
            trace_id=trace_id,
        )

    @app.exception_handler(Exception)
    async def handle_unexpected_error(request: Request, exc: Exception):
        trace_id = _trace_id_from_request(request)
        # Log full traceback, but do not leak details in prod
        logger.exception("Unhandled error on %s", request.url.path)
        return problem_response(
            status=500,
            title="Internal Server Error",
            detail=(
                "Something went wrong. Please contact support."
                if IS_PROD
                else "".join(traceback.format_exception(exc))
            ),
            code="INTERNAL_ERROR",
            instance=str(request.url),
            trace_id=trace_id,
        )



from __future__ import annotations

from typing import Literal, Optional, cast

from fastapi import Body, Depends, Header, Request, Response, status
from starlette.responses import JSONResponse

from svc_infra.apf_payments.schemas import (
    BalanceSnapshotOut,
    CaptureIn,
    CustomerOut,
    CustomerUpsertIn,
    DisputeOut,
    IntentCreateIn,
    IntentListFilter,
    IntentOut,
    InvoiceCreateIn,
    InvoiceLineItemIn,
    InvoiceOut,
    InvoicesListFilter,
    PaymentMethodAttachIn,
    PaymentMethodOut,
    PayoutOut,
    PriceCreateIn,
    PriceOut,
    ProductCreateIn,
    ProductOut,
    RefundIn,
    SetupIntentOut,
    StatementRow,
    SubscriptionCreateIn,
    SubscriptionOut,
    SubscriptionUpdateIn,
    TransactionRow,
    UsageRecordIn,
)
from svc_infra.apf_payments.service import PaymentsService
from svc_infra.api.fastapi.db.sql.session import SqlSessionDep
from svc_infra.api.fastapi.dual import protected_router, public_router, service_router, user_router
from svc_infra.api.fastapi.dual.router import DualAPIRouter
from svc_infra.api.fastapi.middleware.idempotency import require_idempotency_key
from svc_infra.api.fastapi.pagination import (
    Paginated,
    cursor_pager,
    cursor_window,
    sort_by,
    use_pagination,
)

_TX_KINDS = {"payment", "refund", "fee", "payout", "capture"}


def _tx_kind(kind: str) -> Literal["payment", "refund", "fee", "payout", "capture"]:
    if kind not in _TX_KINDS:
        raise ValueError(f"Unknown ledger kind: {kind!r}")
    return cast(Literal["payment", "refund", "fee", "payout", "capture"], kind)


# --- deps ---
async def get_service(session: SqlSessionDep) -> PaymentsService:
    return PaymentsService(session=session)


# --- routers grouped by auth posture (same prefix is fine; FastAPI merges) ---
def build_payments_routers(prefix: str = "/payments") -> list[DualAPIRouter]:
    routers: list[DualAPIRouter] = []

    user = user_router(prefix=prefix, tags=["payments"])
    svc = service_router(prefix=prefix, tags=["payments"])
    prot = protected_router(prefix=prefix, tags=["payments"])

    @user.post(
        "/customers",
        response_model=CustomerOut,
        name="payments_upsert_customer",
    )
    async def upsert_customer(data: CustomerUpsertIn, svc: PaymentsService = Depends(get_service)):
        out = await svc.ensure_customer(data)
        await svc.session.flush()
        return out

    @user.post(
        "/intents",
        response_model=IntentOut,
        name="payments_create_intent",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_intent(
        data: IntentCreateIn,
        request: Request,
        response: Response,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.create_intent(user_id=None, data=data)
        await svc.session.flush()
        response.headers["Location"] = str(
            request.url_for("payments_get_intent", provider_intent_id=out.provider_intent_id)
        )
        return out

    routers.append(user)

    @prot.post(
        "/intents/{provider_intent_id}/confirm",
        response_model=IntentOut,
        name="payments_confirm_intent",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def confirm_intent(provider_intent_id: str, svc: PaymentsService = Depends(get_service)):
        out = await svc.confirm_intent(provider_intent_id)
        await svc.session.flush()
        return out

    @prot.post(
        "/intents/{provider_intent_id}/cancel",
        response_model=IntentOut,
        name="payments_cancel_intent",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def cancel_intent(provider_intent_id: str, svc: PaymentsService = Depends(get_service)):
        out = await svc.cancel_intent(provider_intent_id)
        await svc.session.flush()
        return out

    @prot.post(
        "/intents/{provider_intent_id}/refund",
        response_model=IntentOut,
        name="payments_refund_intent",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def refund_intent(
        provider_intent_id: str, data: RefundIn, svc: PaymentsService = Depends(get_service)
    ):
        out = await svc.refund(provider_intent_id, data)
        await svc.session.flush()
        return out

    @prot.get(
        "/transactions",
        response_model=Paginated[TransactionRow],
        name="payments_list_transactions",
        dependencies=[Depends(cursor_pager(default_limit=50, max_limit=200))],
    )
    async def list_transactions(svc: PaymentsService = Depends(get_service)):
        from sqlalchemy import select

        from svc_infra.apf_payments.models import LedgerEntry

        rows = (await svc.session.execute(select(LedgerEntry))).scalars().all()
        rows_sorted = sort_by(rows, key=lambda e: e.ts, desc=True)

        ctx = use_pagination()
        window, next_cursor = cursor_window(
            rows_sorted,
            cursor=ctx.cursor,
            limit=ctx.limit,
            key=lambda e: int(e.ts.timestamp()),
            descending=True,
        )

        items = [
            TransactionRow(
                id=e.id,
                ts=e.ts.isoformat(),
                type=_tx_kind(e.kind),
                amount=int(e.amount),
                currency=e.currency,
                status=e.status,
                provider=e.provider,
                provider_ref=e.provider_ref or "",
                user_id=e.user_id,
            )
            for e in window
        ]
        return ctx.wrap(items, next_cursor=next_cursor)

    routers.append(prot)

    # PUBLIC webhooks
    pub = public_router(prefix=prefix, tags=["payments"])

    @pub.post("/webhooks/{provider}", name="payments_webhook")
    async def webhooks(
        provider: str,
        request: Request,
        svc: PaymentsService = Depends(get_service),
        signature: Optional[str] = Header(None, alias="Stripe-Signature"),
    ):
        payload = await request.body()
        out = await svc.handle_webhook(provider.lower(), signature, payload)
        await svc.session.flush()
        return JSONResponse(out)

    @user.post(
        "/methods/attach",
        response_model=PaymentMethodOut,
        name="payments_attach_method",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def attach_method(
        data: PaymentMethodAttachIn, svc: PaymentsService = Depends(get_service)
    ):
        out = await svc.attach_payment_method(data)
        await svc.session.flush()
        return out

    @prot.get(
        "/methods",
        response_model=Paginated[PaymentMethodOut],
        name="payments_list_methods",
        dependencies=[Depends(cursor_pager(default_limit=50, max_limit=200))],
    )
    async def list_methods(
        customer_provider_id: str,
        svc: PaymentsService = Depends(get_service),
    ):
        methods = await svc.list_payment_methods(customer_provider_id)
        methods_sorted = sort_by(
            sort_by(methods, key=lambda m: m.provider_method_id or "", desc=False),
            key=lambda m: m.is_default,
            desc=True,
        )
        ctx = use_pagination()
        window, next_cursor = cursor_window(
            methods_sorted,
            cursor=ctx.cursor,
            limit=ctx.limit,
            key=lambda m: m.provider_method_id or "",
            descending=False,
        )
        return ctx.wrap(window, next_cursor=next_cursor)

    @prot.post(
        "/methods/{provider_method_id}/detach",
        name="payments_detach_method",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def detach_method(provider_method_id: str, svc: PaymentsService = Depends(get_service)):
        await svc.detach_payment_method(provider_method_id)
        await svc.session.flush()
        return {"ok": True}

    @prot.post(
        "/methods/{provider_method_id}/default",
        name="payments_set_default_method",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def set_default_method(
        provider_method_id: str,
        customer_provider_id: str,
        svc: PaymentsService = Depends(get_service),
    ):
        await svc.set_default_payment_method(customer_provider_id, provider_method_id)
        await svc.session.flush()
        return {"ok": True}

    # PRODUCTS/PRICES
    @svc.post(
        "/products",
        response_model=ProductOut,
        name="payments_create_product",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_product(data: ProductCreateIn, svc: PaymentsService = Depends(get_service)):
        out = await svc.create_product(data)
        await svc.session.flush()
        return out

    @svc.post(
        "/prices",
        response_model=PriceOut,
        name="payments_create_price",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_price(data: PriceCreateIn, svc: PaymentsService = Depends(get_service)):
        out = await svc.create_price(data)
        await svc.session.flush()
        return out

    # SUBSCRIPTIONS
    @prot.post(
        "/subscriptions",
        response_model=SubscriptionOut,
        name="payments_create_subscription",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_subscription(
        data: SubscriptionCreateIn, svc: PaymentsService = Depends(get_service)
    ):
        out = await svc.create_subscription(data)
        await svc.session.flush()
        return out

    @prot.post(
        "/subscriptions/{provider_subscription_id}",
        response_model=SubscriptionOut,
        name="payments_update_subscription",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def update_subscription(
        provider_subscription_id: str,
        data: SubscriptionUpdateIn,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.update_subscription(provider_subscription_id, data)
        await svc.session.flush()
        return out

    @prot.post(
        "/subscriptions/{provider_subscription_id}/cancel",
        response_model=SubscriptionOut,
        name="payments_cancel_subscription",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def cancel_subscription(
        provider_subscription_id: str,
        at_period_end: bool = True,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.cancel_subscription(provider_subscription_id, at_period_end)
        await svc.session.flush()
        return out

    # INVOICES
    @prot.post(
        "/invoices",
        response_model=InvoiceOut,
        name="payments_create_invoice",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_invoice(
        data: InvoiceCreateIn,
        request: Request,
        response: Response,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.create_invoice(data)
        await svc.session.flush()
        response.headers["Location"] = str(
            request.url_for("payments_get_invoice", provider_invoice_id=out.provider_invoice_id)
        )
        return out

    @prot.post(
        "/invoices/{provider_invoice_id}/finalize",
        response_model=InvoiceOut,
        name="payments_finalize_invoice",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def finalize_invoice(
        provider_invoice_id: str, svc: PaymentsService = Depends(get_service)
    ):
        out = await svc.finalize_invoice(provider_invoice_id)
        await svc.session.flush()
        return out

    @prot.post(
        "/invoices/{provider_invoice_id}/void",
        response_model=InvoiceOut,
        name="payments_void_invoice",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def void_invoice(provider_invoice_id: str, svc: PaymentsService = Depends(get_service)):
        out = await svc.void_invoice(provider_invoice_id)
        await svc.session.flush()
        return out

    @prot.post(
        "/invoices/{provider_invoice_id}/pay",
        response_model=InvoiceOut,
        name="payments_pay_invoice",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def pay_invoice(provider_invoice_id: str, svc: PaymentsService = Depends(get_service)):
        out = await svc.pay_invoice(provider_invoice_id)
        await svc.session.flush()
        return out

    # INTENTS: get/hydrate
    @prot.get("/intents/{provider_intent_id}", response_model=IntentOut, name="payments_get_intent")
    async def get_intent(provider_intent_id: str, svc: PaymentsService = Depends(get_service)):
        return await svc.get_intent(provider_intent_id)

    # STATEMENTS (rollup)
    @svc.get(
        "/statements/daily", response_model=list[StatementRow], name="payments_daily_statements"
    )
    async def daily_statements(
        date_from: str | None = None,
        date_to: str | None = None,
        svc: PaymentsService = Depends(get_service),
    ):
        return await svc.daily_statements_rollup(date_from, date_to)

    # ===== Intents: capture & list =====
    @prot.post(
        "/intents/{provider_intent_id}/capture",
        response_model=IntentOut,
        name="payments_capture_intent",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def capture_intent(
        provider_intent_id: str,
        data: CaptureIn,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.capture_intent(provider_intent_id, data)
        await svc.session.flush()
        return out

    @prot.get(
        "/intents",
        response_model=Paginated[IntentOut],
        name="payments_list_intents",
        dependencies=[Depends(cursor_pager(default_limit=50, max_limit=200))],
    )
    async def list_intents_endpoint(
        customer_provider_id: Optional[str] = None,
        status: Optional[str] = None,
        svc: PaymentsService = Depends(get_service),
    ):
        ctx = use_pagination()
        items, next_cursor = await svc.list_intents(
            IntentListFilter(
                customer_provider_id=customer_provider_id,
                status=status,
                limit=ctx.limit,
                cursor=ctx.cursor,
            )
        )
        return ctx.wrap(items, next_cursor=next_cursor)

    # ===== Invoices: lines/list/get/preview =====
    @prot.post(
        "/invoices/{provider_invoice_id}/lines",
        name="payments_add_invoice_line_item",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def add_invoice_line(
        provider_invoice_id: str,
        data: InvoiceLineItemIn,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.add_invoice_line_item(provider_invoice_id, data)
        await svc.session.flush()
        return {"ok": True, **out}

    @prot.get(
        "/invoices",
        response_model=Paginated[InvoiceOut],
        name="payments_list_invoices",
        dependencies=[Depends(cursor_pager(default_limit=50, max_limit=200))],
    )
    async def list_invoices_endpoint(
        customer_provider_id: Optional[str] = None,
        status: Optional[str] = None,
        svc: PaymentsService = Depends(get_service),
    ):
        ctx = use_pagination()
        items, next_cursor = await svc.list_invoices(
            InvoicesListFilter(
                customer_provider_id=customer_provider_id,
                status=status,
                limit=ctx.limit,
                cursor=ctx.cursor,
            )
        )
        return ctx.wrap(items, next_cursor=next_cursor)

    @prot.get(
        "/invoices/{provider_invoice_id}", response_model=InvoiceOut, name="payments_get_invoice"
    )
    async def get_invoice_endpoint(
        provider_invoice_id: str, svc: PaymentsService = Depends(get_service)
    ):
        return await svc.get_invoice(provider_invoice_id)

    @prot.post(
        "/invoices/preview",
        response_model=InvoiceOut,
        name="payments_preview_invoice",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def preview_invoice_endpoint(
        customer_provider_id: str,
        subscription_id: Optional[str] = None,
        svc: PaymentsService = Depends(get_service),
    ):
        return await svc.preview_invoice(customer_provider_id, subscription_id)

    # ===== Metered usage =====
    @prot.post(
        "/usage_records",
        name="payments_create_usage_record",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_usage_record_endpoint(
        data: UsageRecordIn, svc: PaymentsService = Depends(get_service)
    ):
        out = await svc.create_usage_record(data)
        await svc.session.flush()
        return out

    # ===== Setup Intents (off-session readiness) =====
    @prot.post(
        "/setup_intents",
        name="payments_create_setup_intent",
        status_code=status.HTTP_201_CREATED,
        dependencies=[Depends(require_idempotency_key)],
    )
    async def create_setup_intent(
        payment_method_types: list[str] = Body(default_factory=lambda: ["card"]),
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.create_setup_intent(payment_method_types)
        await svc.session.flush()
        return out

    @prot.post(
        "/setup_intents/{provider_setup_intent_id}/confirm",
        name="payments_confirm_setup_intent",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def confirm_setup_intent(
        provider_setup_intent_id: str, svc: PaymentsService = Depends(get_service)
    ):
        out = await svc.confirm_setup_intent(provider_setup_intent_id)
        await svc.session.flush()
        return out

    @prot.get(
        "/setup_intents/{provider_setup_intent_id}",
        name="payments_get_setup_intent",
        response_model=SetupIntentOut,
    )
    async def get_setup_intent(
        provider_setup_intent_id: str, svc: PaymentsService = Depends(get_service)
    ):
        return await svc.get_setup_intent(provider_setup_intent_id)

    # ===== 3DS/SCA resume (post-action) =====
    @prot.post(
        "/intents/{provider_intent_id}/resume",
        name="payments_resume_intent",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def resume_intent(
        provider_intent_id: str,
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.resume_intent_after_action(provider_intent_id)
        await svc.session.flush()
        return out

    # ===== Disputes =====
    @svc.get(
        "/disputes",
        name="payments_list_disputes",
        response_model=Paginated[DisputeOut],
        dependencies=[Depends(cursor_pager(default_limit=50, max_limit=200))],
    )
    async def list_disputes(
        status: Optional[str] = None,
        svc: PaymentsService = Depends(get_service),
    ):
        ctx = use_pagination()
        items, next_cursor = await svc.list_disputes(
            status=status, limit=ctx.limit, cursor=ctx.cursor
        )
        return ctx.wrap(items, next_cursor=next_cursor)

    @svc.get(
        "/disputes/{provider_dispute_id}",
        name="payments_get_dispute",
        response_model=DisputeOut,
    )
    async def get_dispute(provider_dispute_id: str, svc: PaymentsService = Depends(get_service)):
        return await svc.get_dispute(provider_dispute_id)

    @svc.post(
        "/disputes/{provider_dispute_id}/submit_evidence",
        name="payments_submit_dispute_evidence",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def submit_dispute_evidence(
        provider_dispute_id: str,
        evidence: dict = Body(..., embed=True),  # free-form evidence blob you validate internally
        svc: PaymentsService = Depends(get_service),
    ):
        out = await svc.submit_dispute_evidence(provider_dispute_id, evidence)
        await svc.session.flush()
        return out

    # ===== Balance & Payouts =====
    @svc.get("/balance", name="payments_get_balance", response_model=BalanceSnapshotOut)
    async def get_balance(svc: PaymentsService = Depends(get_service)):
        return await svc.get_balance_snapshot()

    @svc.get(
        "/payouts",
        name="payments_list_payouts",
        response_model=Paginated[PayoutOut],
        dependencies=[Depends(cursor_pager(default_limit=50, max_limit=200))],
    )
    async def list_payouts(svc: PaymentsService = Depends(get_service)):
        ctx = use_pagination()
        items, next_cursor = await svc.list_payouts(limit=ctx.limit, cursor=ctx.cursor)
        return ctx.wrap(items, next_cursor=next_cursor)

    @svc.get(
        "/payouts/{provider_payout_id}",
        name="payments_get_payout",
        response_model=PayoutOut,
    )
    async def get_payout(provider_payout_id: str, svc: PaymentsService = Depends(get_service)):
        return await svc.get_payout(provider_payout_id)

    # ===== Webhook replay (operational) =====
    @svc.post(
        "/webhooks/replay",
        name="payments_replay_webhooks",
        dependencies=[Depends(require_idempotency_key)],
    )
    async def replay_webhooks(
        since: Optional[str] = None,
        until: Optional[str] = None,
        event_ids: Optional[list[str]] = Body(default=None),
        svc: PaymentsService = Depends(get_service),
    ):
        count = await svc.replay_webhooks(since, until, event_ids or [])
        await svc.session.flush()
        return {"replayed": count}

    routers.append(svc)
    routers.append(pub)
    return routers
